# -*- coding: utf-8 -*-
"""
Created on Wed Jun 21 06:31:19 2023

@author: eirik
"""

import math
import random
import pandas as pd
import numpy as np





#CONSTANTS

df = pd.read_excel('results.xlsx', sheet_name='Input')

# Create a dictionary with variable names as keys and their corresponding values as values
fixed_values_constants = {row['Variable']: row['Value'] for _, row in df.iterrows()}

# Replace the fixed values and constants in your code with the values from the dictionary
A_p = fixed_values_constants['A_p'] # area of lead nut
A_s = fixed_values_constants['A_s'] #Valve stem area
A_vb = fixed_values_constants['A_vb'] #Well bore area
A_gs = fixed_values_constants['A_gs'] #Gate seat area
mu = fixed_values_constants['mu'] #Friction coeff, seat and gate
F_f = fixed_values_constants['F_f'] #Stem seal friction force
P_w = fixed_values_constants['P_w'] #Water pressure
L = fixed_values_constants['L'] #Lead screw ft/rev
eff = fixed_values_constants['eff'] #Lead screw efficiency
G_R = fixed_values_constants['G_R'] #Gear ratio
E_1 = fixed_values_constants['E_1'] #Motor efficiency
E_2 = fixed_values_constants['E_2'] #Gear efficiency
rpm_open = fixed_values_constants['rpm_open']  # RPM during opening.
rpm_close = fixed_values_constants['rpm_close']  # RPM during closing.
k = fixed_values_constants['k'] #flow consistency index
n = fixed_values_constants['n'] #flow behaviour index
dP = fixed_values_constants['dP_test'] #Differential pressure set for leakage test
test_type=fixed_values_constants['Test Type'] #Normal or Partial for type of valve testing

#VALVE INTERFACE 

# Set up empty list for torque data
force_data = []
torque_data = []


# Calculate the range for WHSIP which decreases and random dP_extra that represents excess dP when opening valve after test. Also set a fixed dP for testing.
WHSIP_range = np.linspace(1800, 1000, num=100)
dP_extra = np.random.uniform(low=0, high=435, size=100)


#Friction decrease base parameters
base_decrease_chance_f_f = 0.2
base_decrease_interval_f_f = (10, 50)

#Loop for simulating linear forces present in each state of the valve cycle. Also simmulates decrease in stem seal friction: magnitude and chance is based on excess dP when opening valve.
for i in range(100):
    WHSIP = int(WHSIP_range[i])
    dP_random = int(dP_extra[i])
    # Calculate the decrease chance and interval for F_f based on D_P
    decrease_chance_f_f = base_decrease_chance_f_f * dP_random / max(dP_extra)
    decrease_interval_f_f = (base_decrease_interval_f_f[0] * dP_random / max(dP_extra),
                             base_decrease_interval_f_f[1] * dP_random / max(dP_extra))

    # Check if F_f should be decreased and update its value
    if random.random() < decrease_chance_f_f:
        F_f -= random.uniform(*decrease_interval_f_f)
        F_f = max(0, F_f)  # Ensure F_f doesn't become negative
    
    # Forces
    F_1 = WHSIP * mu * (2*A_gs + A_vb)  # Force due to seat gate friction
    F_2 = F_f                           # Force due to seal friction
    F_3 = P_w * (A_p-A_s)               # Force on piston from water pressure
    F_4 = WHSIP * A_s                   # Force on lower stem from well shut in pressure 
   

    # step 1: interval before pressure equalization
    F_t1 = F_1 + F_2 + F_3 + F_4
    
    # step 2: interval from crack open to fully open
    F_t2 = F_2 + F_3 + F_4
    
    # step 3: interval from fully open to crack close
    F_t3 = -F_2 + F_3 + F_4
    
    # step 4: interval after pressure build up
    F_t4 = -F_1 - F_2 + F_3 + F_4 
    
    # add force data to list
    force_data.append([WHSIP, dP, dP_random, F_t1, F_t2, F_t3, F_t4, F_f])
    # convert data to pandas dataframe 
    df_force = pd.DataFrame(force_data, columns=['WHSIP','dP','dP_random', 'F_t1', 'F_t2', 'F_t3', 'F_t4', 'F_f'])



#LEAD SCREW

# New loop to convert from linear forces to torques

# eff represents the efficiency of the lead screw and decreases randomly of given chance within the given interval
eff_decrease_interval = (0.0000005, 0.0000015)
eff_decreasechance=0.1
# n represents the flow behaviour index of the fluid around the lead screw and increases randomly of given chance within the given interval
n_increase_interval = (0.000005, 0.00015)
n_increasechance=0.1
for i in range(100):
    if random.random() < eff_decreasechance:  
        eff_decrease = random.uniform(*eff_decrease_interval)
        eff -= eff_decrease
        
    if random.random() < n_increasechance:
        n_increase = random.uniform(*n_increase_interval)
        n += n_increase
        
    WHSIP, dP, dP_random, F_t1, F_t2, F_t3, F_t4, F_f = force_data[i]

    T_t1 = F_t1 * L / (2 * math.pi * eff) * (1 + k * (rpm_open ** (n-1)))
    T_t2 = F_t2 * L / (2 * math.pi * eff) * (1 + k * (rpm_open ** (n-1)))
    T_t3 = F_t3 * L / (2 * math.pi * eff) * (1 + k * (rpm_close ** (n-1)))
    T_t4 = F_t4 * L / (2 * math.pi * eff) * (1 + k * (rpm_close ** (n-1)))
    
    torque_data.append([WHSIP, dP, dP_random, T_t1, T_t2, T_t3, T_t4, eff, n])

df_torque = pd.DataFrame(torque_data, columns=['WHSIP', 'dP', 'dP_random', 'T_t1', 'T_t2', 'T_t3', 'T_t4', 'eff','n'])




#MOTOR AND GEAR

# set up empty lists for power data
power_t1 = []
power_t2 = []
power_t3 = []
power_t4 = []



# loop through torque data and calculate power for each torque value. This is to simulate power needed as motor input
for row in df_torque.itertuples(index=False):

    # extract torque values
    tau_t1 = row[3]
    tau_t2 = row[4]
    tau_t3 = row[5]
    tau_t4 = row[6]
    
    # calculate motor output torque and power input for each torque value
    tau_motor_t1 = tau_t1 / (G_R * E_2)
    power_t1.append((tau_motor_t1 * rpm_open) / (E_1 * 5252))
    
    tau_motor_t2 = tau_t2 / (G_R * E_2)
    power_t2.append((tau_motor_t2 * rpm_open) / (E_1 * 5252))
    
    tau_motor_t3 = tau_t3 / (G_R * E_2)
    power_t3.append((tau_motor_t3 * rpm_close) / (E_1 * 5252))
    
    tau_motor_t4 = tau_t4 / (G_R * E_2)
    power_t4.append((tau_motor_t4 * rpm_close) / (E_1 * 5252))

# create dictionary of power data with column names
power_data = {'WHSIP': df_torque['WHSIP'], 'dP': df_torque['dP'], 'dP_random': df_torque['dP_random'], 'power_t1': power_t1, 'power_t2': power_t2, 'power_t3': power_t3, 'power_t4': power_t4}



#CONVERT TO TIMESERIES

#Set time intervals based on rpm, Gear ratio and ft/rev
distance_24mm = 24 * 0.00328084                         # approximately 0.07874 feet for crack
distance_196mm = 196 * 0.00328084                       # approximately 0.64304 feet from crack to end
crack_open_time = distance_24mm/(rpm_open/G_R*L)*60     # Time used for crack open
run_open_time = distance_196mm/(rpm_open/G_R*L)*60      # Time used for crack open
stay_open_time = 100                                    # Time used for crack open
run_close_time = distance_196mm/(rpm_close/G_R*L)*60    # Time used for crack open
crack_close_time = distance_24mm/(rpm_close/G_R*L)*60   # Time used for crack open
stay_close_time = 100                                   # Time used for crack open
total_time = round(run_close_time + crack_close_time + stay_close_time + crack_open_time + run_open_time + stay_open_time)
total_time_partial = round(run_close_time*0.2 + stay_close_time + run_open_time*0.2 + stay_open_time)



# Main timeseries loop

# Assigns WHSIP and calculated force, torque and power data to timeseries (seconds and cycles). Also assigns the state, dP, rpm and position on premisses. Additional data for eff, n and f_f is included for verification in excel.
leakage = 0
dP_random_prev = None
#Cycles through calculated data with length 100
for cycle in range(100):
    
    #Defines a dataframe to store generated data
    
    # Collect torque data
    WHSIP, dP, dP_random, T_t1, T_t2, T_t3, T_t4, eff, n = torque_data[cycle]
    # 7th index of force_data contains F_f
    F_f = force_data[cycle][7]    
    
    #Loop for normal valve testing
    
    if test_type == 'Normal':
        df_timeseries_cycle = pd.DataFrame(index=range(total_time), columns=["Timestamps [s]", "Cycle", "State", "WHSIP [psi]", "dP [psi]", "Position [mm]", "RPM", "Linear Force [lbf]", "Power [HP]", "Torque [lbf*ft]", "Lead Screw eff", "Flow beh.index, n", "Seal Friction [lbf]"])
        for t in range(total_time):
            if t < run_close_time:  
                state = "run_close"
                dP_current = 0
                force = df_force['F_t3'][cycle]
                power = power_t3[cycle]
                torque = T_t3
                rpm = rpm_close
                position = -220 + ((220 - 24) / run_close_time) * t                                                                             # Position changes linearly from -220 to -24
            elif t < run_close_time + crack_close_time:  
                state = "crack_close"
                dP_current = 0
                force = df_force['F_t4'][cycle]
                power = power_t4[cycle]
                torque = T_t4
                rpm = rpm_close
                position = -24 + (24 / crack_close_time) * (t - run_close_time)                                                                 # Position changes linearly from -24 to 0
            elif t < run_close_time + crack_close_time + stay_close_time:  
                state = "stay_closed"
                force = df_force['F_t1'][cycle]
                power = power_t1[cycle]
                torque = T_t1
                rpm = 0
                position = 0  # Valve stays closed
                if t < run_close_time + crack_close_time + 10:
                    dP_current = 0
                elif t < run_close_time + crack_close_time + 20:
                    dP_current = dP * (t - run_close_time - crack_close_time - 10) / 10  # Gradually increase to testing dP over 10 seconds
                elif t < run_close_time + crack_close_time + 35:
                    # Start decreasing dP with potential leakage after 15 seconds
                    dP_current = dP - leakage * (t - run_close_time - crack_close_time - 15) / 15
                else:
                    # Decrease dP to dP_random = excess dP over next 10 seconds
                    dP_current = max((dP - leakage * 15) - (dP - leakage * 15 - dP_random) * (t - run_close_time - crack_close_time - 35) / 10, dP_random)
                    # dP decreases in preparation for opening. Normaly after 10 minutes, but here 15 sec.
    
            elif t < run_close_time + crack_close_time + stay_close_time + crack_open_time:  
                state = "crack_open"
                force = df_force['F_t1'][cycle]
                power = power_t1[cycle]
                torque = T_t1
                rpm = rpm_open
                position = -24 * ((t - run_close_time - crack_close_time - stay_close_time) / crack_open_time)                                  # Position changes linearly from 0 to -24
                dP_current = dP_random                                                                                                          # Excess dP remains for crack open
            elif t < run_close_time + crack_close_time + stay_close_time + crack_open_time + run_open_time:  
                state = "run_open"
                dP_current = 0
                force = df_force['F_t2'][cycle]
                power = power_t2[cycle]
                torque = T_t2
                rpm = rpm_open
                position = -24 - ((220 - 24) / run_open_time) * (t - run_close_time - crack_close_time - stay_close_time - crack_open_time)     # Position changes linearly from -24 to -220       
            else:  
                state = "stay_open"
                dP_current = 0
                force = df_force['F_t3'][cycle]
                power = power_t3[cycle]
                torque = T_t3
                rpm = 0
                position = -220                                                                                                                  # Valve stays open             
    
            df_timeseries_cycle.loc[t] = [t + cycle * total_time, cycle, state, WHSIP, dP_current, position, rpm, force,  power, torque, eff, n , F_f]
              
            #Generates potential leakage based on the excess dP from previous cycle
        if dP_random_prev is not None:
            if np.random.rand() < (dP_random_prev/10): # Adjust the denominator for a suitable chance
                leakage_increase = np.random.rand() * dP_random_prev * 0.01 # Adjust the multiplier for a suitable increase magnitude
                leakage = min(max(leakage + leakage_increase, 0), dP_random) # Ensure leakage does not exceed dP_random_prev
        dP_random_prev = dP_random
    
        if cycle == 0:
            df_timeseries = df_timeseries_cycle.copy()
        else:
            df_timeseries = pd.concat([df_timeseries, df_timeseries_cycle])
            
    #Loop for partial stroke testing ---> Does not include crack close and crack open. Position and time is ajusted for only 20% of run close       
    elif test_type == 'Partial':
       df_timeseries_cycle = pd.DataFrame(index=range(total_time_partial), columns=["Timestamps [s]", "Cycle", "State", "WHSIP [psi]", "dP [psi]", "Position [mm]", "RPM", "Linear Force [lbf]", "Power [HP]", "Torque [lbf*ft]", "Lead Screw eff", "Flow beh.index, n", "Seal Friction [lbf]"])
       for t in range(total_time_partial):
           end_position_run_close = -220+(220-24)*0.2                                               # end position for run_close state in partial stroke test
           if t < run_close_time*0.2:  
               state = "run_close"
               dP_current = 0
               force = df_force['F_t3'][cycle]
               power = power_t3[cycle]
               torque = T_t3
               rpm = rpm_close
               position = -220 + ((end_position_run_close + 220) / (run_close_time*0.2)) * t       # Position changes linearly from -220 to end_position_run_close over the time 'run_close_time*0.2'                                                                   # Position changes linearly from -220 to -180.4 = (-220+(220-24)*0.2)
                                                                          
           elif t < run_close_time*0.2 + stay_close_time:  
               state = "stay_closed"
               dP_current = 0
               force = df_force['F_t2'][cycle]
               power = power_t2[cycle]
               torque = T_t2
               rpm = 0
               position = end_position_run_close                                                   # Valve stays at end_position_run_close
               
   
           elif t < run_close_time*0.2 + stay_close_time + run_open_time*0.2:  
               state = "run_open"
               dP_current = 0
               force = df_force['F_t2'][cycle]
               power = power_t2[cycle]
               torque = T_t2
               rpm = rpm_open
               position = end_position_run_close - ((end_position_run_close + 220) / (run_open_time*0.2)) * (t - run_close_time*0.2 - stay_close_time) # Position changes linearly from end_position_run_close to -220 over the time 'run_open_time*0.2'      
                                                                                                                        
           else: 
               state = "stay_open"
               dP_current = 0
               force = df_force['F_t3'][cycle]
               power = power_t3[cycle]
               torque = T_t3
               rpm = rpm_open
               position = -220                                                                      # Valve stays open at -220
           df_timeseries_cycle.loc[t] = [t + cycle * total_time_partial, cycle, state, WHSIP, dP_current, position, rpm, force, power, torque, eff, n , F_f]
       if cycle == 0:
           df_timeseries = df_timeseries_cycle.copy()
       else:
           df_timeseries = pd.concat([df_timeseries, df_timeseries_cycle])
df_timeseries.reset_index(drop=True, inplace=True)



# Save to Excel
df_timeseries.to_excel("timeseries.xlsx")

# Write the updated DataFrame back to the Excel file
try:
    with pd.ExcelWriter('results.xlsx', mode='a', engine='openpyxl', if_sheet_exists='overlay') as writer:
        df_timeseries.to_excel(writer, sheet_name='Data', index=False)
    
    
except FileNotFoundError:
    with pd.ExcelWriter('results.xlsx', mode='w', engine='openpyxl') as writer:
        df_timeseries.to_excel(writer, sheet_name='Data', index=False)


